/* ///////////////////////////////////////////////////////////////////////////
//  MC = < Manifold Code >
//  Copyright (C) 1994--2000  Michael Holst
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  Free Software Foundation; either version 2 of the License, or (at your
//  option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  675 Mass Ave, Cambridge, MA 02139, USA.
//
//  rcsid="$Id: main.c,v 1.1 2003/03/08 00:13:48 fetk Exp $"
// //////////////////////////////////////////////////////////////////////// */
#include "smol.h"

VEMBED(rcsid="$Id: main.c,v 1.4 2003/07/09 15:49:34 fetk Exp $")
VPRIVATE void ARG_setParam(char **argv, int argc, Vlparam *lparam, Vtparam *tparam) 
{
    int i;
   
    /* Default Values */
    sprintf(lparam->ifnam,"smol.in");        /* input file name */
    sprintf(lparam->ofnam,"outputNUM.dx");   /* output file name */
    sprintf(lparam->afnam,"analout.dx");     /* analytical output */
    sprintf(tparam->rstnam,"rst.m");         /* restart file */
    sprintf(lparam->Sfnam,"outputNUMS.dx");  /* local entropy density */
    sprintf(lparam->Gfnam,"outputNUMG.dx");  /* local free energy density */

    lparam->prob  = 0;               /* index for primal or dual problem */
    lparam->lkey  = 2;               /* iterative method */
    lparam->lmax  = 5000;            /* Max no. iterative steps*/
    lparam->gues  = 0;               /* Guess Type */
    lparam->prec  = 1;               /* Preconditioner type */
    lparam->ptype = 3;               /* Output type */
    lparam->ltol  = 1e-6;            /* Tolerance of iterative solver */
    lparam->pjac  = 2;               /* index for printing the system matrix */
    lparam->key   = 0;               /* smooth with A or A' (0=A, 1=A') */
    lparam->flag  = 0;               /* determines which "mode" we run in */
    lparam->cycle = 0;
 
    tparam->tkey    = 1;             /* Type of time-dependent problems */
    tparam->nstep   = 10000;         /* toal time steps */
    tparam->pfreq   = 1000;          /* frequency to write restart files */
    tparam->efreq   = 1;             /* frequency to write energy files */
    tparam->ekeytol = 4;             /* total number of requested energies */
    tparam->pdetype = 1;             /* type of PDEs */
    tparam->dt      = 0.0005;        /* time step of time-dependent problems */
    tparam->t0      = 0.;            /* initial time of time-dependent problems */
    tparam->cfreq   = 100;           /* frequency to write time-dependent density files */

    for (i=1; i<argc; i+=2) {
        if (!strcmp(argv[i],"-ifnam"))  sprintf(lparam->ifnam,"%s",argv[i+1]);
        if (!strcmp(argv[i],"-ofnam"))  sprintf(lparam->ofnam,"%s",argv[i+1]);
        if (!strcmp(argv[i],"-rstnam"))  sprintf(tparam->rstnam,"%s",argv[i+1]);
        if (!strcmp(argv[i],"-Sfnam"))  sprintf(lparam->Sfnam,"%s",argv[i+1]);
        if (!strcmp(argv[i],"-Gfnam"))  sprintf(lparam->Gfnam,"%s",argv[i+1]);
        if (!strcmp(argv[i],"-lkey"))   lparam->lkey    = atoi(argv[i+1]);          
        if (!strcmp(argv[i],"-lmax"))   lparam->lmax    = atoi(argv[i+1]);          
        if (!strcmp(argv[i],"-ltol"))   lparam->ltol    = atof(argv[i+1]);          
        if (!strcmp(argv[i],"-prec"))   lparam->prec    = atoi(argv[i+1]);          
        if (!strcmp(argv[i],"-gues"))   lparam->gues    = atoi(argv[i+1]);          
        if (!strcmp(argv[i],"-ptype"))  lparam->ptype   = atoi(argv[i+1]);
        if (!strcmp(argv[i],"-tkey"))   tparam->tkey    = atoi(argv[i+1]);
        if (!strcmp(argv[i],"-nstep"))  tparam->nstep   = atoi(argv[i+1]);
        if (!strcmp(argv[i],"-pfreq"))  tparam->pfreq   = atoi(argv[i+1]);
        if (!strcmp(argv[i],"-efreq"))  tparam->efreq   = atoi(argv[i+1]);
        if (!strcmp(argv[i],"-cfreq"))  tparam->cfreq   = atoi(argv[i+1]);
        if (!strcmp(argv[i],"-pde"))    tparam->pdetype = atoi(argv[i+1]);
        if (!strcmp(argv[i],"-dt"))     tparam->dt      = atof(argv[i+1]);
        if (!strcmp(argv[i],"-t0"))     tparam->t0      = atof(argv[i+1]);
    }    
}                           
int main(int argc, char **argv) {

    Gem *gm         = VNULL;
    PDE *pde        = VNULL;
    AM *am          = VNULL;
    Aprx *aprx      = VNULL;
    Vio *sock       = VNULL;
    Valist *alist   = VNULL;
    Vmgrid *mgrid   = VNULL; 
    Vmodel *model   = VNULL;
    Vlparam *lparam = VNULL;
    Vtparam *tparam = VNULL;
    Vgrid *grid[50];
    int i, ip[20], ngrid, isolve, marked, maxsolve=1;
    int nsimps, nverts;
    int bytesTotal, highWaterTotal;
    double rp[20], energies[20];
    VV *vx;

    /* Start the I/O layer */
    Vio_start();
    
    /* Parse command line flags */ 
    lparam = lparam_ctor();
    tparam = tparam_ctor();
    ARG_setParam( argv, argc, lparam, tparam);

    /* Construct all related objects */
    pde   = myPDE_ctor();
    gm    = Gem_ctor(VNULL, pde);
    aprx  = Aprx_ctor(VNULL, gm, pde);
    am    = AM_ctor(VNULL, aprx);
    model = usr_getModel(pde);

    /* Read input file */                                    
    ngrid = PDE_input(am, grid, lparam);
    Mesh_ActiveSite(am);

    /* Solving PDE */
    Vnm_print(1,"Starting SMOL solver...\n");

    switch (model->tmkey) {
        case 0:
             break;
        case 1:
             for (isolve=0; isolve<maxsolve; isolve++) {
                 Vnm_print(1,"\nmain: Solve-estimate-refine #%d.\n", isolve);

                 nsimps = Gem_numSS(gm);
                 nverts = Gem_numVV(gm);
                 Vnm_print(1,"main: Have %d simplices and %d vertices TOTAL\n", nsimps, nverts);

                 Gem_reorderSV(gm);
                 Vnm_redirect(0);
                 Gem_shapeChk(gm);
                 Vnm_redirect(1);

                 dPMF_mapping(am);
//                 for (vx=Gem_firstVV(gm);vx!=VNULL;vx=Gem_nextVV(gm)) {
//                     Vnm_print(0,"model->PMF[%ld] = %le\n",VV_id(vx),model->PMF[VV_id(vx)]);
//                 }
                 MY_lSolve(am, lparam);

                 if (isolve != (maxsolve-1)) {
                    Vnm_print(1,"main: Refining...\n");
                    marked = AM_markRefine(am, 2, -1, 1, 1000000.0);
                    if (marked == 0) {
                       Vnm_print(1, "main: No simplices marked for refinement; breaking!\n");
                    break;
                    }
                    AM_refine(am, 0, 0, 0);
                    Vnm_print(1,"main: Marked %d simps, have %d after refinement\n", marked, Gem_numSS(gm));
                 }
             }
             break;
        case 2:
             nsimps = Gem_numSS(gm);
             nverts = Gem_numVV(gm);
             Vnm_print(1,"main: Have %d simplices and %d vertices TOTAL\n", nsimps, nverts);

             Gem_reorderSV(gm);
             Vnm_redirect(0);
             Gem_shapeChk(gm);
             Vnm_redirect(1);

             dPMF_mapping(am);
             MY_tSolve(am, lparam, tparam);
             break;
        default:
             nsimps = Gem_numSS(gm);
             nverts = Gem_numVV(gm);
             Vnm_print(1,"main: Have %d simplices and %d vertices TOTAL\n", nsimps, nverts);

             Gem_reorderSV(gm);
             Vnm_redirect(0);
             Gem_shapeChk(gm);
             Vnm_redirect(1);

             dPMF_mapping(am);
             MY_lSolve(am, lparam);
             break;
        }

        Vnm_redirect(0);
        Vmem_print(VNULL);
        Vmem_printTotal();
        Vnm_redirect(1);
        bytesTotal = Vmem_bytesTotal();
        highWaterTotal    = Vmem_highWaterTotal();
        Vnm_print(1,"main: %f MB TOTAL\n",  bytesTotal/(1024.0*1024.0));
        Vnm_print(1,"main: %f MB HIGH WATER\n", highWaterTotal/(1024.0*1024.0));

    sprintf(lparam->ofnam,"outputNUM.dx");
    MY_write(am, lparam, energies);

    if (model->diffmesh) {
       sock=Vio_ctor("FILE", "ASC", "localhost", "adaptive_mesh.m", "w");
       Gem_write(gm, 0, sock, 0);
       Vio_dtor(&sock);
    }

    AM_dtor(&am);
    Aprx_dtor(&aprx);
    Gem_dtor(&gm);
    myPDE_dtor(pde);

    Vnm_print(2,"\n");
    Vnm_print(2,"mc_leaks = [\n");
    Vnm_print(2,"%% --------------------------------------"
                "--------------------------------------\n");
    Vnm_print(2,"%%  Footprint        Areas       Malloc         Free"
                "    Highwater   Class\n"),
    Vnm_print(2,"%% --------------------------------------"
                "--------------------------------------\n");
    Vmem_print(VNULL);
    Vmem_printTotal();
    Vnm_print(2,"%% --------------------------------------"
                "--------------------------------------\n");
    Vnm_print(2,"];\n");

    return 0;
}

