/* ///////////////////////////////////////////////////////////////////////////
//  MC = < Manifold Code >
//  Copyright (C) 1994--2001  Michael Holst and co-authors
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  Free Software Foundation; either version 2 of the License, or (at your
//  option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  675 Mass Ave, Cambridge, MA 02139, USA.
//
//  rcsid="$Id: main.c,v 1.16 2002/12/20 03:06:42 fetk Exp $"
// //////////////////////////////////////////////////////////////////////// */

/*
 * ***************************************************************************
 * File:     myroutines.c
 *
 * Purpose:  Collection of modified routines from mc and mcx.
 *
 * Author:   Jason Suen
 * ***************************************************************************
 */

#include "smol.h"

#define COULOM_COEF(TEMP,CHG,RAD,COEF) COEF = 1.67149e+5*CHG/(80*TEMP*RAD)
#undef _DEBUG
#define _ANALYTIC
#define len_w 15
#undef _LOG_OF_U
VPRIVATE Bvec *w[len_w];
VPRIVATE int winuse;
VPRIVATE int iline[4][3];

VPUBLIC void MY_assemR(AM *thee, Bvec *u){
    int i, inode, smid, face, key, qp;
    double jacD, Dw, theta;
    Gem *gm  = thee->aprx->gm;
    Re *re = thee->aprx->re[0];
    SS *sm;
    TT t;

    smid = 0;
    while ( (smid < Gem_numSS(gm)) && (!Vnm_sigInt()) ) {
        sm = Gem_SS(gm,smid);
        Gem_buildVolumeTrans(gm,sm,&t);
        face=-1; key=0; jacD=VABS(t.D);
        for (qp=0; qp<Re_numQ_hi(re,face); qp++){
            Dw = jacD*Re_w_hi(re,qp,face);
            for (i=0; i<Gem_dimVV(gm); i++){
                inode = t.vid[i];
                theta = Dw*Re_phi_hi(re,qp,i,face);
                Bvec_addToB(u, 0, inode, theta);
            }
        }
        smid++;
    }
}

VPUBLIC void MY_assem(AM *thee, int evalKey, Bvec *u, Bvec *f, int ip[], double rp[],
    double energies[], int ekeytol) {
    int smid, i, j, k, iv, r, s, inode, jnode, face, qp,key, tset, fType;
    int dim, dimVV, numSS, numQ;
    double jacD, Dw, theta, theta2, xq[3], phi[4], phix[4][3], U[MAXV], dU[MAXV][3], nvec[3];
    PDE *pde = thee->aprx->pde;
    Gem *gm  = thee->aprx->gm;
    Re *re = thee->aprx->re[0], *reU = thee->aprx->re[0];
    SS *sm;
    Emat *em;
    TT t;

    dim = Gem_dim(gm);
    dimVV = Gem_dimVV(gm);
    numSS = Gem_numSS(gm);

    for (i=0; i<dim; i++){
        xq[i] = 0.;
        dU[0][i] = 0.;
        for (j=0; j<dimVV; j++) phix[j][i]=0.;
    }

//    Vnm_print(2,"evalKey = %d \n",evalKey);
//    Vnm_print(2,"    Key = %d \n",key);

    pde->initAssemble(pde,ip,rp);
    for (i=0; i<ekeytol; i++) energies[i] = 0;
    for (smid=0; smid<numSS; smid++){
        sm = Gem_SS(gm,smid);

        /* volume trans from master to this element (and back) */
        Gem_buildVolumeTrans(gm,sm,&t);

        /* once-per-element PDE initialization */
        pde->initElement(pde,t.stype,t.gchart,t.vx,&t);

        face=-1;   // Volumn
        for (qp=0; qp<Re_numQ_hi(re,face); qp++) {

            /* jacobian and quadrature weight */
            Dw = t.D * Re_w_hi(re,qp,face);

           /* build basis functions */
            for(i=0; i<3; i++) nvec[i] = t.nvec[face][i]; 
            Aprx_buildPhi(thee->aprx,re,qp,face,&t,phi,phix);
            Aprx_buildU(thee->aprx,reU,evalKey,qp,face,&t,U,dU,thee->u,thee->ud);
            Aprx_buildXq(thee->aprx,re,qp,face,&t,xq);
            pde->initPoint(pde,key,t.gchart,xq,U,dU);

           /* assemble energy if required */
            energies[2] += Dw*pde->Ju(pde,2);
            energies[3] += Dw*pde->Ju(pde,3);
        }

        // surface
        for (face=0; face<dimVV; face++){
            fType = SS_faceType(sm,face);
            t.focusFace = face;
            if (fType == IB_RD){
                Gem_buildSurfaceTrans(gm,face,&t);
                pde->initFace(pde,fType,t.gchart,t.nvec[face]);
                jacD = VABS(t.faceD[face]);
                key = 1;
                for (qp=0; qp<Re_numQ_hi(re,face); qp++){
                    for(i=0; i<3; i++) nvec[i] = t.nvec[face][i];
                    Dw = jacD*Re_w_hi(re,qp,face);
                    Aprx_buildPhi(thee->aprx,re,qp,face,&t,phi,phix);
                    Aprx_buildU(thee->aprx,reU,1,qp,face,&t,U,dU,thee->u,thee->ud);
                    Aprx_buildXq(thee->aprx,re,qp,face,&t,xq);
                    pde->initPoint(pde,key,t.gchart,xq,U,dU);
                    for (i=0; i<ekeytol; i++) energies[i] += Dw*pde->Ju(pde,ip[i]);
                }
            }
        }
    }

}

VPUBLIC int MY_Solve(AM *am, Bvec *u, Bvec *w[], Vlparam *lparam){

    int prob=lparam->prob, lkey=lparam->lkey, lmax=lparam->lmax;
    int gues=lparam->gues, prec=lparam->prec, pjac=lparam->pjac;
    int key =lparam->key, flag=lparam->flag, cycle=lparam->cycle;
    int iter;
    double ltol=lparam->ltol;
    Bmat *A=am->A, *P=am->P;
    Bvec *f=am->f, *r=am->r, *ut=am->ut;

    /* method choice (0=slu,1=mg,2=cg,3=bcg,4=bcg,5=bcgstab,6=continuation) */
    switch (lparam->lkey) {

        /* primary methods */
//        case 0:  Bvec_slu( u, A, f, r, ut, key, flag);
//                 break;

        case 1:  Bvec_mg( u, A, f, r, ut, key, flag, lmax, ltol, prec,
                     cycle, P);
                 break;
        case 2:  iter = MY_cg(u,A,f,r,ut,key,flag,lmax,ltol,prec,cycle,P,w);
                 break;
        case 3:  iter = MY_cgT(u,A,f,r,ut,key,flag,lmax,ltol,prec,cycle,P,w);
                 break;
        case 4:  Bvec_bcg( u, A, f, r, ut, key, flag, lmax, ltol, prec,
                      cycle, P,
                      w[0],w[1],w[2],w[3],w[4],w[5],w[6],w[7],w[8],w[9],w[10]);
                 break;
        case 5:  iter = MY_bcg(u,A,f,r,ut,key,flag,lmax,ltol,prec,cycle,P,w);
                 break;
        case 6:  iter = MY_bcgstab(u,A,f,r,ut,key,flag,lmax,ltol,prec,cycle,P,w);
                 break;
        default:
                 Vnm_print(0,"Bvec_lSolve: Bad meth given!\n");
                 break;
    }

    return iter;
}
/*
 * ***************************************************************************
 * Routine:  MY_lSolve
 *
 * Purpose:  Modified linear solver.
 *
 * Author:   Jason Suen, followed Michael Holst
 * ***************************************************************************
 */
VPUBLIC void MY_lSolve(AM *thee, Vlparam *lparam)
{
    /* assembly parameters */
    int key, flag, cycle, ip[10], iter;
    int evalKey, energyKey, residKey, tangKey, massKey;
    int prob=lparam->prob, lkey=lparam->lkey, lmax=lparam->lmax;
    int gues=lparam->gues, prec=lparam->prec, pjac=lparam->pjac;
    double ltol=lparam->ltol, rp[10];
    Bmat *A = thee->A, *P = thee->P;
    Bvec *u_use;
    Bvec *u = thee->u, *ud = thee->ud, *ut = thee->ut;
    Bvec *w0 = thee->w0, *f=thee->f, *r=thee->r; 

    int i, myNverts;
    double Pot, xc, yc, zc, xyz;
    VV *vxSMOL;
    Vec *conc, *concBD;

    /* some i/o */
    Vnm_print(0,"AM_lSolve: starting..");

    /* make sure we have structure */
    AM_create(thee);

    /* setup for assembling either primal or dual problem */
    energyKey=0; massKey=0; evalKey= 2*prob; residKey=1; tangKey=1;

    if (prob == 0) {       /* forward (primal) problem (linearized about 0) */
        u_use     = u;  
    } else if (prob == 1){ /* backward (dual) problem (linearized about 0) */
        u_use     = w0; 
    } else { VASSERT(0); }

    /* initial guess */
    if (gues == 1) {
        Vnm_print(0,"..using zero as initial approximation..\n");
        Bvec_init(u_use, 0.);
        AM_applyDiriZero( thee, u_use );
    } else {
        Vnm_print(0,"..using u_D as initial approximation..\n");
        AM_iniGuess( thee, u_use );
        AM_applyDiriZero( thee, u_use );
    }

    /* clear the matrix, rhs, residual */
    Bmat_zero( A ); Bvec_init( f,  0. ); Bvec_init( r,  0. );

    /* assemble the stiffness matrix and load vector */
    AM_assem(thee,evalKey,energyKey,residKey,tangKey,massKey,0,u,ud,f,ip,rp);

    Vnm_print(0,"Starting iterative solver \n");
    if (lkey>1) Bvec_createVectors(u,w,len_w);
    (void *)MY_Solve(thee,u_use,w,lparam);

    if (lkey>1) Bvec_destroyVectors(u,w,len_w);

    Vnm_print(0,"AM_lSolve: done.\n");
    /* return with no errors */
    return;
}

VPUBLIC void MY_write(AM *thee, Vlparam *lparam, double *energies){
 
    int i, j, ip[20], ekeytol=4; 
    double rp[20];
    double analrad1, analrad2, analrate, coef1, coef2;
    Vio *sock=VNULL;
    Bmat *A=thee->A, *M=thee->M;
    Bvec *f=thee->f, *u=thee->u, *ud=thee->ud, *ut=thee->ut;
    Bvec *resid, *w0=thee->w0, *ui=thee->ui;
    PDE *pde = thee->aprx->pde;
    Vmodel *model  = usr_getModel(pde);
    Gem *gm = thee->aprx->gm;
    SS *sm;
    VV *vx;
    TT t;
    double expp,u_tmp,upot,iconc0,entrT,entr,fe_tmp;

    for (i=0; i<Gem_numVV(gm); i++) {
        expp = VEXP(-(model->PMF)[i]);
        u_tmp = expp*Bvec_valB(u,0,i) + Bvec_valB(ud,0,i);
        Bvec_setB(ut,0,i,u_tmp);
        if (model->entropy||model->gibbs) {
           iconc0 = u_tmp/model->conc;
           upot = 0.001*u_tmp*(model->PMF)[i]/model->charge;
           if (iconc0 > VSMALL) {
              entrT = - u_tmp*VLOG(iconc0);  // unit of entroy is cal/mol
              entr  = entrT/model->T;
              fe_tmp = upot - entrT*0.001;    // unit of free energy is kcal/mol
           } else {
              entrT = 0;
              entr  = 0;
              fe_tmp = upot;         
           }
           Bvec_setB(w0,0,i,entr);
           Bvec_setB(ui,0,i,fe_tmp);
        }
    }

    sock = Vio_ctor("FILE", "ASC", "localhost", lparam->ofnam,"w");
    switch ( lparam->ptype) {
        case 1:  Aprx_writeGEOM(thee->aprx, sock, 1,1,-1,1,1, ut,"GV");
                 break;
        case 2:  Aprx_writeSOL( thee->aprx, sock, ut, "GMV" );
             break;
        case 3:  Aprx_writeSOL( thee->aprx, sock, ut, "DX" );
                 break;
        case 4:  Aprx_writeSOL( thee->aprx, sock, ut, "UCD" );
                 break;
        case 5:  Aprx_writeSOL( thee->aprx, sock, ut, "TEC" );
                 break;
        default:
             Vnm_print(2,"Alg_writeSVio: unsupported print type.\n");
         break;
    }
    Vio_dtor(&sock);

    if (model->entropy) {
       Bvec_scal(w0,kT_kcal);

       sock = Vio_ctor("FILE", "ASC", "localhost", lparam->Sfnam,"w");
       switch ( lparam->ptype) {
           case 1:  Aprx_writeGEOM(thee->aprx, sock, 1,1,-1,1,1, w0,"GV");
                    break;
           case 2:  Aprx_writeSOL( thee->aprx, sock, w0, "GMV" );
                    break;
           case 3:  Aprx_writeSOL( thee->aprx, sock, w0, "DX" );
                    break;
           case 4:  Aprx_writeSOL( thee->aprx, sock, w0, "UCD" );
                    break;
           case 5:  Aprx_writeSOL( thee->aprx, sock, w0, "TEC" );
                    break;
           default:
                Vnm_print(2,"Alg_writeSVio: unsupported print type.\n");
                break;
       }
       Vio_dtor(&sock);
    }

    if (model->gibbs) {
       Bvec_scal(ui,kT_kcal);

       sock = Vio_ctor("FILE", "ASC", "localhost", lparam->Gfnam,"w");
       switch ( lparam->ptype) {
           case 1:  Aprx_writeGEOM(thee->aprx, sock, 1,1,-1,1,1, ui,"GV");
                    break;
           case 2:  Aprx_writeSOL( thee->aprx, sock, ui, "GMV" );
                    break;
           case 3:  Aprx_writeSOL( thee->aprx, sock, ui, "DX" );
                    break;
           case 4:  Aprx_writeSOL( thee->aprx, sock, ui, "UCD" );
                    break;
           case 5:  Aprx_writeSOL( thee->aprx, sock, ui, "TEC" );
                    break;
           default:
                Vnm_print(2,"Alg_writeSVio: unsupported print type.\n");
                break;
       }
       Vio_dtor(&sock);
    }

#ifdef _ANALYTIC
    analrad1 = 10.0;
    analrad2 = 200.0;
    if (VABS(model->charge) > VSMALL) {
       COULOM_COEF(300,model->charge,analrad1,coef1);
       COULOM_COEF(300,model->charge,analrad2,coef2);
       analrate = 3.612e4*4*VPI*model->diff*coef1*analrad1/(exp(coef1 - coef2)- 1.0);
    } else {
       analrate = 3.612e2*4*VPI*model->diff*VSQR(analrad1)/(1/analrad1 - 1/analrad2);
    }
    Vnm_print(0,"diff  : %g\n", model->diff);
    Vnm_print(0,"TEMP  : %g\n", 300.0);
    Vnm_print(0,"Analytical kinetic constant: %g M^-1 min^-1.\n", analrate);
#endif

    if (model->tmkey == 1) {
       for (i=0;i<ekeytol;i++) ip[i] = i;
       MY_assem(thee,1,u,f,ip,rp,energies,ekeytol);
       energies[0] = energies[0]/model->conc;
       Vnm_print(0, "Rxn rate through area integral is %g A^3/us; or = %g M^-1 min^-1. \n"
                 , energies[0], energies[0]*6e7*ang3_iM);
       Vnm_print(0, "Surface area for the reactive boundary is %e A^2. \n", energies[1]);
       Vnm_print(0, "Total substrate is %e. \n", energies[2]/1660.6);
       Vnm_print(0, "Total volume is %e. \n", energies[3]);
       Vnm_tstop(41,"main");
    }
}

VPRIVATE void PQR_input(AM *thee, char path[]) {

    PDE *pde = thee->aprx->pde;
    Vmodel *model = usr_getModel(pde);
    Vio *sock=VNULL;

    sock = Vio_ctor("FILE", "ASC", "localhost", path, "r");
//    model->alist = Valist_ctor();
    Valist_readPQR(model->alist, VNULL, sock);
    Vio_dtor(&sock);
}

VPRIVATE void PDE_readMesh(AM *am, char path[]){
    Vio *sock;

    sock = Vio_ctor("FILE", "ASC", "localhost", path, "r");
    AM_read(am,0,sock);
    Vio_dtor(&sock);
}

VPRIVATE int Mesh_isActive(AM *thee, VV *vx)
{
    PDE *pde = thee->aprx->pde;
    Vmodel *model = usr_getModel(pde);
    Gem *gm = thee->aprx->gm;
    int i;
    double xq[3],distx,disty,distz,dist;

    for (i=0;i<Gem_dim(gm);i++) xq[i] = VV_coord(vx,i);
    for (i=0;i<model->nsp;i++) {
        distx = VABS(xq[0]-model->ctx[i]);
        disty = VABS(xq[1]-model->cty[i]);
        distz = VABS(xq[2]-model->ctz[i]);
        if ((distx<model->rad[i]) && (disty<model->rad[i]) && (distz<model->rad[i])) {
           dist = VSQRT(VSQR(distx) + VSQR(disty) + VSQR(distz));
           if (dist < model->rad[i]) return model->color[i];
        }
    }

    return 0;
}

VPUBLIC void Mesh_ActiveSite(AM *thee) 
{
    PDE *pde = thee->aprx->pde;
    Vmodel *model = usr_getModel(pde);
    Gem *gm = thee->aprx->gm;
    SS *sm;
    int i, j, vv, color;

    for (sm=Gem_lastSS(gm); sm!=VNULL; sm=Gem_prevSS(gm)) {
        for (i=0; i<Gem_dimVV(gm); i++) {
            if ((SS_faceType(sm,i)==4)) {
               for (j=0; j<Gem_dimVV(gm)-1; j++) {
                   int vv = SS_faceVertexNumber(sm, i, j);
                   color = Mesh_isActive(thee,SS_vertex(sm,vv));
                   if (color) {
                      VV_setType( SS_vertex(sm,vv), color );
                      SS_setFaceType( sm, i, color );
                   }
               }
            }
        }
    }
}

VPRIVATE void Mesh_readActive(AM *thee, char path[]){
    Vio *sock;
    
    sock = Vio_ctor("FILE", "ASC", "localhost", path, "r");
    MY_readActive(thee,sock);
    Vio_dtor(&sock);
}

VPUBLIC int MY_readActive(AM *thee, Vio *sock){

    char tok[VMAX_BUFSIZE];
    char *MCwhiteChars = " =,;\t\n", *MCcommChars  = "#%";
    PDE *pde = thee->aprx->pde;
    Vmodel *model = usr_getModel(pde);
    int i=0,color;
    double rad, ctx, cty, ctz;

    /* Setup comment and whitespace characters */
    Vio_setWhiteChars(sock, MCwhiteChars);
    Vio_setCommChars(sock, MCcommChars);

    while (Vio_scanf(sock, "%s", tok) == 1) {
        if (Vstring_strcasecmp(tok, "rcent") == 0) {
           VJMPERR2( Vio_scanf(sock,"%le",&rad) );
           VJMPERR2( Vio_scanf(sock,"%le",&ctx) );
           VJMPERR2( Vio_scanf(sock,"%le",&cty) );
           VJMPERR2( Vio_scanf(sock,"%le",&ctz) );
           VJMPERR2( Vio_scanf(sock,"%d",&color) );
           model->rad[i] = rad;
           model->ctx[i] = ctx;
           model->cty[i] = cty;
           model->ctz[i] = ctz;
           model->color[i] = color;
        }
        i++;
    }
    model->nsp = i;
    Vnm_print(0, "Done reading active site file.\n");
    return 1;

VERROR2:
    Vnm_print(2,"MY_readActive: Format problem with input\n");
    return 0;
}

VPUBLIC void dPMF_mapping(AM *am) {

    PDE *pde = am->aprx->pde;
    Gem *gm  = am->aprx->gm;
    Vmodel *model = usr_getModel(pde);
    Vmgrid  *mgrid  = model->mgrid;
    VV *vx;
    double xq[3], value, dPot[3], curv[3], Pot, tcurv;
    int i;

    model->PMF = Vmem_malloc(gm->vmem, Gem_numVV(gm), sizeof(double));
    VASSERT(model->PMF != VNULL);
    for (i=0; i<Gem_numVV(gm); i++) (model->PMF)[i] = 0;

    (model->dPMF)[0] = (void *)Vmem_malloc(gm->vmem, Gem_numVV(gm), sizeof(double));
    VASSERT((model->dPMF)[0] != VNULL);
    for (i=0; i<Gem_numVV(gm); i++) model->dPMF[0][i] = 0;

    (model->dPMF)[1] = (void *)Vmem_malloc(gm->vmem, Gem_numVV(gm), sizeof(double));
    VASSERT((model->dPMF)[1] != VNULL);
    for (i=0; i<Gem_numVV(gm); i++) model->dPMF[1][i] = 0;

    (model->dPMF)[2] = (void *)Vmem_malloc(gm->vmem, Gem_numVV(gm), sizeof(double));
    VASSERT((model->dPMF)[2] != VNULL);
    for (i=0; i<Gem_numVV(gm); i++) model->dPMF[2][i] = 0;

    for (vx=Gem_firstVV(gm);vx!=VNULL;vx=Gem_nextVV(gm)) {
        for (i=0;i<3;i++) xq[i]=VV_coord(vx,i);
        switch (model->map) {
           case 0:
                if (!Vmgrid_value(mgrid, xq, &Pot)){
                   Vnm_print(2, " Vmgrid_value failed!\n");
                   VASSERT(0);
                }
                if (!Vmgrid_gradient(mgrid, xq, dPot)){
                   Vnm_print(2, " Vmgrid_gradient failed!\n");
                   VASSERT(0);
                }
                break;
           case 1:
                if (!Vmgrid_eval(mgrid, xq, &Pot, dPot, curv)){
                   Vnm_print(2, " Vmgrid_gradient failed!\n");
                   VASSERT(0);
                }
                break;
           default:
                if (!Vmgrid_value(mgrid, xq, &Pot)){
                   Vnm_print(2, " Vmgrid_value failed!\n");
                   VASSERT(0);
                }
                if (!Vmgrid_gradient(mgrid, xq, dPot)){
                   Vnm_print(2, " Vmgrid_gradient failed!\n");
                   VASSERT(0);
                }
        }
        (model->PMF)[VV_id(vx)] = model->charge*Pot;
        (model->dPMF)[0][VV_id(vx)] = model->charge*dPot[0];
        (model->dPMF)[1][VV_id(vx)] = model->charge*dPot[1];
        (model->dPMF)[2][VV_id(vx)] = model->charge*dPot[2];
    }
}

VPUBLIC int PDE_input(AM *thee, Vgrid *grid[], Vlparam *lparam)
{
    Vio *sock=VNULL;
    char *MCwhiteChars = " =,;\t\n", *MCcommChars  = "#%";
    char cdum[80], path[VMAX_ARGLEN];
    int i, idum, ngrid=0;                                              
    double dum, F[3];                                         
    FILE *chartfile;
    VV *vx;
    int myNverts;
    float p1;
   
    Vnm_print(0, "\nReading input file (%s)...\n", lparam->ifnam);
    sock = Vio_ctor("FILE", "ASC", "localhost", lparam->ifnam, "r");

    PDE *pde = thee->aprx->pde;
    Gem *gm  = thee->aprx->gm;

    Vmodel *model  = usr_getModel(pde);

    Valist *alist = model->alist;
    Vmgrid *mgrid = model->mgrid;

    /* Setup comment and whitespace characters */             
    Vio_setWhiteChars(sock, MCwhiteChars);                    
    Vio_setCommChars(sock, MCcommChars);                      
                                                              
    Vnm_print(0, "Start reading input file parsing...\n");         
    while (!Vnm_sigInt()) {                                   
                                                              
        if (Vio_scanf(sock, "%s", cdum) != 1) {               
            break;                                            
        }
        if (!strcmp(cdum,"Diff_coef")){
            VJMPERR2( 1==Vio_scanf(sock,"%le",&dum) );
            model->diff = dum;
        }
        if (!strcmp(cdum,"sub_charge")){
            VJMPERR2( Vio_scanf(sock,"%le",&dum) );
            model->charge = dum;
        }
        if (!strcmp(cdum,"sub_conc")){
            VJMPERR2( 1==Vio_scanf(sock,"%le",&dum) );
            model->conc = dum;
        }
        if (!strcmp(cdum,"temp")){
            VJMPERR2( 1==Vio_scanf(sock,"%le",&dum) );
            model->T = dum;
        }
        if (!strcmp(cdum,"tmkey")){
            VJMPERR2( 1==Vio_scanf(sock,"%s",cdum) );
            if (!strcmp(cdum,"NONE")) model->tmkey = 0;
            else if (!strcmp(cdum,"SSSE")) model->tmkey = 1;
            else if (!strcmp(cdum,"TDSE")) model->tmkey = 2;
            else model->tmkey = 0;
        }
        if (!strcmp(cdum,"map")){
            VJMPERR2( 1==Vio_scanf(sock,"%s",cdum) );
            if (!strcmp(cdum,"NONE")) model->map = 0;
            else if (!strcmp(cdum,"DIRECT")) model->map = 0;
            else if (!strcmp(cdum,"FEM")) model->map = 1;
            else model->map = 0;
        }
        if (!strcmp(cdum,"Diff_Mesh")){
            VJMPERR2( 1==Vio_scanf(sock,"%s",cdum) );
            if (!strcmp(cdum,"NO")) model->diffmesh = 0;
            else if (!strcmp(cdum,"YES")) model->diffmesh = 1;
            else model->diffmesh = 0;
        }
        if (!strcmp(cdum,"Entropy")){
            VJMPERR2( 1==Vio_scanf(sock,"%s",cdum) );
            if (!strcmp(cdum,"NO")) model->entropy = 0;
            else if (!strcmp(cdum,"YES")) model->entropy = 1;
            else model->entropy = 0;
        }
        if (!strcmp(cdum,"Gibbs")){
            VJMPERR2( 1==Vio_scanf(sock,"%s",cdum) );
            if (!strcmp(cdum,"NO")) model->gibbs = 0;
            else if (!strcmp(cdum,"YES")) model->gibbs = 1;
            else model->gibbs = 0;
        }
        /* Input for I/O information */                        
        if (!strcmp(cdum,"mol")){
            VJMPERR2(Vio_scanf(sock, "%s", path) == 1);
            PQR_input(thee, path);
            Vnm_print(0,"Finish reading molecule \n");
        }
        if (!strcmp(cdum,"mesh")){
            VJMPERR2(Vio_scanf(sock, "%s", path) == 1);
            PDE_readMesh(thee, path);
            Vnm_print(0,"Finish reading mesh \n");
        }
        if (!strcmp(cdum,"mgrid")){                   
            VJMPERR2(Vio_scanf(sock, "%s", path) == 1);
            grid[ngrid]=Vgrid_ctor(0, 0, 0, 0., 0., 0., 0., 0., 0., VNULL);
            VASSERT(Vgrid_readDX(grid[ngrid], "FILE", "ASC", VNULL, path));
            Vnm_print(0, "Finish reading grid from %s. \n",path);
            Vmgrid_addGrid(model->mgrid, grid[ngrid]); 
            ngrid++;
        }
        if (!strcmp(cdum,"active")){
            VJMPERR2(Vio_scanf(sock, "%s", path) == 1);
            Mesh_readActive(thee, path);
            Vnm_print(0,"Finish reading mesh active site \n");
        } 
        if (!strcmp(cdum,"end")){ break; }
    }                                                         

    Vnm_print(0, "Done reading input file.\n");               
    Vio_dtor(&sock);                                                          
    return ngrid;                                                 
                                                              
VERROR2:                                                      
    Vnm_print(2,"MY_read: Format problem with input\n");  
    Vio_dtor(&sock);
    return ngrid;                                                 
}

#include "mysolv.c"
#include "tsolve.c"
VPUBLIC Vlparam* lparam_ctor(){
    Vlparam *thee;
    thee = Vmem_malloc( VNULL, 1, sizeof(Vlparam) );
    return thee;
}

VPUBLIC void lparam_dtor(Vlparam *thee){
    Vmem_free(VNULL, 1, sizeof(Vlparam), (void **)(&(thee)));
}

