/* ///////////////////////////////////////////////////////////////////////////
//  MC = < Manifold Code >
//  Copyright (C) 1994--2001  Michael Holst and co-authors
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  Free Software Foundation; either version 2 of the License, or (at your
//  option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  675 Mass Ave, Cambridge, MA 02139, USA.
//
//  rcsid="$Id: main.c,v 1.16 2002/12/20 03:06:42 fetk Exp $"
// //////////////////////////////////////////////////////////////////////// */

/*
 * ***************************************************************************
 * File:     mysolv.c
 *
 * Purpose:  Collection of modified solvers
 *
 * Author:   Jason Suen
 * ***************************************************************************
 */
VPRIVATE int prtKey, iters;
VPRIVATE double denom,rsnrm,mytol;
VPRIVATE double alpha, beta, rho, omega;
VPRIVATE char rn[80], rmeth[80], pmeth[80];

/*
 * ***************************************************************************
 * Routine:  Bvec_nrm2
 *
 * Purpose:  2-norm of a block vector.
 *
 * Author:   Yuhui Cheng
 * ***************************************************************************
 */
VPRIVATE double Bvec_nrm2B(Bvec *thee)
{   int i;
    double tmp = 0.0;
    Vec *uu;

    for (i=0; i<Bvec_numB(thee); i++) { 
        uu = thee->VC[i];   
        tmp += Vec_nrm2( (Vec*)uu );
    }
    return tmp;    
}

VPRIVATE void Vec_axpby(Vec *thee, Vec *s, double valx, double valy)
{
    int i, n, m, remain;
    double *u, *v;

    n = thee->n;
    m = s->n;
    VASSERT( n == m );
    u = thee->u;
    v = s->u;

    remain = (n%4);
    for (i=0; i<remain; i++) u[i] = valy*u[i] + valx*v[i];
    for (i=remain; i<n; i+=4){
        u[i]   = valy*u[i]   + valx*v[i];
        u[i+1] = valy*u[i+1] + valx*v[i+1];
        u[i+2] = valy*u[i+2] + valx*v[i+2];
        u[i+3] = valy*u[i+3] + valx*v[i+3];
    }
}

VPRIVATE void Bvec_axpby(Bvec *thee, Bvec *v, double valx, double valy){
    int i;

    VASSERT( Bvec_numRT(thee) == Bvec_numRT(v) );

    for (i=0; i<thee->numB; i++) Vec_axpby( thee->VC[i], v->VC[i], valx, valy);

}

VPRIVATE int MY_initSolver(Bvec *r, Bmat *A, Bvec *u, Bvec *f, 
                    int prec, int flag, double etol)
{
    int value;

    if (prec == 1) {
        strncpy(pmeth,"(ds):",80);
    } else if (prec == 2) {
        strncpy(pmeth,"(jac):",80);
    } else if (prec == 3) {
        strncpy(pmeth,"(mg):",80);
    } else {
        strncpy(rmeth,":",80);
    }
    if (flag==2) {
        sprintf(rn,"--->%s%s",rmeth,pmeth);
        prtKey = 50;
    } else {
        sprintf(rn,"%s%s",rmeth,pmeth);
        prtKey = 1;
    }

    /* iteration parameters */
    alpha    = 0.0;
    beta     = 1.0;
    iters    = 0;
          
    /* DO: r = f - A u, rtilde = r*/
    Bvec_matvec( r, A, u, 0, 0);
    Bvec_axpy( r, f, (-1.0) );
    Bvec_scal( r, (-1.0) );

    rsnrm = Bvec_nrm2(r);
    denom = rsnrm;
    mytol  = etol*denom;
    value = (rsnrm < VSMALL);
    if (value) {
        Vnm_print(0,"%sinitial residual=%8.2e\n", rn, rsnrm);
        Vnm_print(0,"%sinitial residual is zero...returning.\n", rn);
        return value;
    }
    if ((flag%2)==0) Vnm_print(0,"%sinitial residual=%8.2e\n", rn, rsnrm);

    return value;
}

VPRIVATE void MY_ICG(Vec *thee, Mat *amat, Vec *f, Vec *w){

    int i, j, k, numR;
    int *IA, *JA;
    double *uu, *ww, *ff, wtmp;
    double *diag, *offU, *offL;

    /* some local variable pointers for speed and readability */
    numR = amat->numR;
    uu   = Vec_addr(thee);
    ww   = Vec_addr(w);
    ff   = Vec_addr(f);
    IA   = amat->IA;
    JA   = amat->JA;
    diag = amat->diag;
    offU = amat->offU;
    offL = amat->offL;

    /* ww = diag * (diag + offU)^-1 ff */
    for (i=numR-1; i>=0; i--){
        wtmp=ff[i];
        for (j=IA[i+1]-1; j>=IA[i]; j--){
            k = JA[j];
            wtmp -= offU[j]*ww[k];
        }
        ww[i] = wtmp;
    }

    /* uu = (diag+offL)^-1 ww */
    for (i=0; i<numR; i++){
        uu[i] = 0.;
        wtmp = ww[i];
        for (j=IA[i]; j<IA[i+1]; j++){
            k = JA[j];
            wtmp -= offL[j]*uu[k];
        } 
        uu[i] = wtmp/diag[i];
    }
}

VPRIVATE int MY_io(Bvec *r, int flag)
{
    int value;

    rsnrm = Bvec_nrm2B(r);
    value = (rsnrm < mytol);
  
    if ((flag%2) == 0){
        if  ( ((iters%prtKey)==0) || value){
            Vnm_print(0,"%s it=%4d",rn,iters);
            Vnm_print(0,"  rr=%8.2e  r=%8.2e\n", rsnrm/denom, rsnrm);
        }
    }
    return value;
}
VPRIVATE void MY_precond(Bvec *thee, Bmat *A, Bvec *f, Bmat *P, 
        Bvec *tmp1, Bvec *tmp2, int prec, int key){

    int i,j,itmax=5,it=0;
    int key_m, flag_m, itmax_m, etol_m, prec_m, cycle_m;

    if (prec == 1){
        Bvec_diagScale(thee,A,f);
    } else if (prec == 2) {
        MY_ICG(thee->VC[0],A->AD[0][0],f->VC[0],tmp1->VC[0]);
    } else if (prec == 3) {
        key_m   = key;
        flag_m  = 1;
        itmax_m = 1;
        etol_m  = 0.;
        prec_m  = prec;
        cycle_m = 2;
        Bvec_mg(thee, A, f, thee, thee,
            key_m, flag_m, itmax_m, etol_m, prec_m, cycle_m, P);
    } else {
        Bvec_copy(thee,f);
    }
}
VPRIVATE void MY_normal(Bvec *thee){
    double normal;

    normal = 1./Bvec_nrm2(thee);
    if (normal>VLARGE) normal = 1;
    Bvec_scal(thee,normal);
}

/*
 * ***************************************************************************
 * Routine:  Bvec_cg
 *
 * Purpose:  CG method (ODIR algorithm, requires SPD operators A, B, and M).
 *
 *           ODIR(M,B,A) (for Au=f, preconditioner B, inner-product M)
 *           ---------------------------------------------------------
 *
 *           Given u^0
 *           r^0 = f - A u^0
 *           p^0 = B r^0 / ||B r^0||
 *           for i=0,1,2,....
 *               alpha_i = <M e^i, p^i> / <M p^i, p^i>
 *               u^{i+1} = u^i + alpha_i p^i
 *               r^{i+1} = r^i - alpha_i A p^i
 *               gamma_i = <M B A p^i, p^i> / <M p^i, p^i>
 *               sigma_i = <M B A p^i, p^{i-1}> / <M p^{i-1}, p^{i-1}>
 *               v       = B A p^i - gamma_i p^i - sigma_i p^{i-1}
 *               p^{i+1} = v / ||v||
 *           end for
 *
 *           CG=ODIR(M=A,B=B,A=A)
 *           --------------------
 *
 *           Given u^0
 *           r^0 = f - A u^0
 *           p^0 = B r^0 / ||B r^0||
 *           for i=0,1,2,....
 *               alpha_i = <r^i, p^i> / <A p^i, p^i>
 *               u^{i+1} = u^i + alpha_i p^i
 *               r^{i+1} = r^i - alpha_i A p^i
 *               gamma_i = <B A p^i, A p^i> / <A p^i, p^i>
 *               sigma_i = <B A p^i, A p^{i-1}> / <A p^{i-1}, p^{i-1}>
 *               v       = B A p^i - gamma_i p^i - sigma_i p^{i-1}
 *               p^{i+1} = v / ||v||
 *           end for
 *
 * Input:    The "flag" variable determines which "mode" we run in:
 *
 *           flag==0 --> Normal: check itmax, error tolerance; normal i/o
 *           flag==1 --> Silent: check only itmax; no i/o
 *           flag==2 --> Subcycle: check itmax, error tolerance; subcycle i/o
 *           flag==3 --> Subcycle: check itmax, error tolerance; no i/o
 *
 * Author:   Michael Holst and Stephen Bond
 * ***************************************************************************
 */
VPUBLIC int MY_cg(Bvec *u, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key,
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[])
{
    double pAp, poApo, BApAp, BApApo, rp, gamma, sigma;
    Bvec *p=w[1], *ap=w[2], *bap=w[3], *po=w[4], *apo=w[5], *tp=w[6];
    Bvec *tmp1=w[7], *tmp2=w[8];

    int prtKey = VPRTKEY;

    strncpy(rmeth,"JKS_cg",80);
    MY_initSolver(r, A, u, f, prec, flag, etol);
    MY_precond(p,A,r,P,tmp1,tmp2,prec,key);

    /* iteration parameters */
    iters   = 0;
    pAp     = 1.;
    poApo   = 1.0;
    BApAp   = 0.0;
    BApApo  = 0.0;
    Bvec_init(apo,0.);

    /* DO: for i=0,1,2,.... */
    while ((iters<itmax) && !Vnm_sigInt()) {

        /* DO: p^{i+1} = v / ||v|| */
        MY_normal(p);

        /* DO: alpha_i = <r^i, p^i> / <A p^i, p^i> */
        Bvec_matvec( ap, A, p, key, 0);
        poApo = pAp;
        pAp   = 1./Bvec_dot( p, ap );
        rp    = Bvec_dot( r, p );
        alpha = rp * pAp;

        /* DO: u^{i+1} = u^i + alpha_i p^i */
        Bvec_axpy( u, p, alpha );

        /* DO: r^{i+1} = r^i - alpha_i A p^i */
        Bvec_axpy( r, ap, (-alpha) );


        /* if the main solver, do some i/o */
        if (MY_io(r, flag)) return iters;

        /* form new direction vector */
        /* DO: (form: B A p^i) */
        MY_precond(bap,A,ap,P,tmp1,tmp2,prec,key);

        /* DO: gamma_i = <B A p^i, A p^i> / <A p^i, p^i> */
        BApAp = Bvec_dot( bap, ap );
        gamma = BApAp * pAp;

        /* DO: sigma_i = <B A p^i, A p^{i-1}> / <A p^{i-1}, p^{i-1}> */
        BApApo = Bvec_dot( bap, apo );
        sigma = BApApo * poApo;

        /* DO: v = B A p^i - gamma_i p^i - sigma_i p^{i-1} */
        Bvec_copy( apo, ap );
        Bvec_axpy( bap, p, (-gamma) );
        Bvec_axpy( bap, po, (-sigma) );
        Bvec_copy( po, p );
        Bvec_copy( p, bap );

        /* iterate */
        iters++;
    }

    if ( (flag==0) || (flag==2) )
        Vnm_print(0,"%s itmax reached! (rr=%e)\n", rn, rsnrm/denom);
    return iters;
}

/*
 * ***************************************************************************
 * Routine:  Bvec_cg
 *
 * Purpose:  CG method (pure bi-orthogonal extension of CG).
 *
 * Notes:    This is a CG implementation based on SIAM template book
 *
 *           ------------------------------------------------------------
 *
 *           Given u
 *           r      = f - A  u
 *           z      = M^{-1} r
 *           rho    = dot(r,z)
 *           for i=0,1,2,...
 *                    q = A* p
 *               alpha  = rho/dot(q, p)
 *                   x += alpha *p
 *                   r -= alpha *q
 *                   z  = M^{-1} r
 *                 rho2 = rho
 *                 rho = (z, r)
 *                 beta = rho1/rho2
 *                    p = z + beta * p
 *           end for
 *
 * Input:    The "flag" variable determines which "mode" we run in:
 *
 *           flag==0 --> Normal: check itmax, error tolerance; normal i/o
 *           flag==1 --> Silent: check only itmax; no i/o
 *           flag==2 --> Subcycle: check itmax, error tolerance; subcycle i/o
 *           flag==3 --> Subcycle: check itmax, error tolerance; no i/o
 *
 * Author:   Jason Suen
 * ***************************************************************************
 */
VPUBLIC int MY_cgT(Bvec *u, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key, 
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[]){

    Bvec *z=w[1], *p=w[2], *q=w[3], *tmp1=w[4], *tmp2=w[5];

    strncpy(rmeth,"JKS_cgT",80);
    if (MY_initSolver(r, A, u, f, prec, flag, etol)) return 0;

    /* pvec = r, qvec = s */
    MY_precond(z,A,r,P,tmp1,tmp2,prec,key);
    rho = Bvec_dot(z,r);
    Bvec_copy(p,z);

    /* DO: for i=0,1,2,.... */
    while ((iters < itmax) && (!Vnm_sigInt())) {

        /* STEP 1: q = A * p */
        Bvec_matvec(q, A, p, key, 0);

        /* STEP 2: alpha = rho / dot (ptilde, q) */
        alpha = rho / Bvec_dot(p, q);

        /* STEP 3: x += alpha * p */
        Bvec_axpy(u, p, alpha);

        /* STEP 4: r -= alpha * q */
        Bvec_axpy(r, q,-alpha);

        /* STEP 5: Update z */
        MY_precond(z, A, r, P, tmp1, tmp2, prec,key);

        /* STEP 6: beta = (r,s)/ (r_n, s_n) */
        beta  = 1.0/rho;
        rho  = Bvec_dot(z,r);
        beta *= rho;

        /* STEP 7: p = z + beta * p */
        Bvec_axpby(p,z,1.0,beta);

        /* STEP 8: calculate norm of current residual, do some i/o */
        if ( MY_io(r,flag) ) return iters;

        /* iterate */
        iters++;
    }

    if ( (flag==0) || (flag==2) )
        Vnm_print(0,"%s itmax reached! (rr=%e)\n", rn, rsnrm/denom);
    return iters;
}
/*
 * ***************************************************************************
 * Routine:  Bvec_bcg
 *
 * Purpose:  BCG method (pure bi-orthogonal extension of CG).
 *
 * Notes:    This is a BCG implementation based on SIAM template book
 *
 *           ------------------------------------------------------------
 *
 *           Given u
 *           r      = f - A  u
 *           rtilde = r
 *           z      = M^{-1} r
 *           ztilde = M^T^{-1} rtilde
 *           for i=0,1,2,...
 *                    q = A* p
 *               qtilde = A^T * p
 *               alpha  = rho/dot(ptilde, q)
 *                   x += alpha *p
 *                   r -= alpha *q
 *               rtilde-= alpha*qtilde
 *                   z  = M^{-1} r
 *               ztilde = M^T^{-1} rtilde
 *                 rho2 = rho1
 *                 rho1 = (z, rtilde)
 *                 beta = rho1/rho2
 *                    p = z + beta * p
 *               ptilde = ztilde + beta * ptilde
 *           end for
 *
 * Input:    The "flag" variable determines which "mode" we run in:
 *
 *           flag==0 --> Normal: check itmax, error tolerance; normal i/o
 *           flag==1 --> Silent: check only itmax; no i/o
 *           flag==2 --> Subcycle: check itmax, error tolerance; subcycle i/o
 *           flag==3 --> Subcycle: check itmax, error tolerance; no i/o
 *
 * Author:   Jason Suen
 * ***************************************************************************
 */
VPUBLIC int MY_bcg(Bvec *u, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key, 
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[]){

    Bvec *rtilde=w[1], *z=w[2], *ztilde=w[3];
    Bvec *p=w[4], *ptilde=w[5], *q=w[6], *qtilde=w[7];
    Bvec *tmp1 = w[8], *tmp2 = w[9];
    int iters2=0;

    strncpy(rmeth,"JKS_bcg",80);
    if (MY_initSolver(r, A, u, f, prec, flag, etol)) return 0;

    /* DO: rtilde = r*/
    Bvec_copy(rtilde,r);

    /* pvec = r, qvec = s */
    MY_precond(z,     A,r,     P, tmp1, tmp2,prec,key);
    MY_precond(ztilde,A,rtilde,P, tmp1, tmp2,prec,!key);
    rho = Bvec_dot(z,rtilde);
    Bvec_copy(p,z);
    Bvec_copy(ptilde,ztilde);

    /* DO: for i=0,1,2,.... */
    while ((iters < itmax) && (!Vnm_sigInt())) {

        /* STEP 1: q = A * p, qtilde = A^T * ptilde */
        Bvec_matvec(q,      A, p,       key, 0);
        Bvec_matvec(qtilde, A, ptilde, !key, 0);

        /* STEP 2: alpha = rho / dot (ptilde, q) */
        alpha = rho / Bvec_dot(ptilde, q);
         if (VABS(alpha)>VLARGE) {
             Vnm_print(0,"%s (Ap,q)=%e (rr=%e)\n", rn, alpha, rsnrm/denom);
             Vnm_print(0,"%s Hitting the roundoff wall...returning.\n",rn);
            return iters;
         } 
        /* STEP 3: x += alpha * p */
        Bvec_axpy(u,p,alpha);

        /* STEP 4: r -= alpha * q; rtilde -=alpha * qtilde  */
        Bvec_axpy(r,      q,      -alpha);
        Bvec_axpy(rtilde, qtilde, -alpha);

        /* STEP 5: Update z and ztilde */
        MY_precond(z,     A,r,     P,tmp1,tmp2,prec,key);
        MY_precond(ztilde,A,rtilde,P,tmp1,tmp2,prec,!key);

        /* STEP 6: beta = (r,s)/ (r_n, s_n) */
        beta  = 1.0/rho;
        rho  = Bvec_dot(z,rtilde);
        beta *= rho;

        /* STEP 7: p = z + beta * p */
        Bvec_axpy(z,p,beta);
        Bvec_copy(p,z);

        /* STEP 8: ptilde = ztilde + beta * ptilde */
        Bvec_axpy(ztilde,ptilde,beta);
        Bvec_copy(ptilde,ztilde);

        /* STEP 9: calculate norm of current residual, do some i/o */
        if ( MY_io(r,flag) ) return iters;

        /* STEP 10: restart solver every 500 steps */
        iters++;
    }

    Vnm_print(0,"%s itmax reached! (rr=%e)\n", rn, rsnrm/denom);
    return iters;
}
/*
 * ***************************************************************************
 * Routine:  Bvec_bcgstab
 *
 * Purpose:  BCGstab method (improved bi-orthogonal extension of CG).
 *
 * Notes:    This is an improved BCG routine. (1) no matrix transpose is
 *           is involved. and (2) the convergence is smoother.
 *
 * Algorithm:
 *     Initialization - r = b - A x, r0 = r, p = r, rho = (r0, r),
 *     Iterate -
 *     (1) v = A p
 *     (2) alpha = rho / (r0, v)
 *     (3) s = r - alpha v
 *     (4) t = A s
 *     (5) omega = (t, s) / (t, t)
 *     (6) x = x + alpha * p + omega * s
 *     (7) r = s - omega * t
 *     convergence test goes here
 *     (8) beta = rho, rho = (r0, r), beta = rho * alpha / (beta * omega)
 *         p = r + beta * (p - omega * v)
 *
 * Input:    The "flag" variable determines which "mode" we run in:
 *
 *           flag==0 --> Normal: check itmax, error tolerance; normal i/o
 *           flag==1 --> Silent: check only itmax; no i/o
 *           flag==2 --> Subcycle: check itmax, error tolerance; subcycle i/o
 *           flag==3 --> Subcycle: check itmax, error tolerance; no i/o
 *
 * Author:   Jason Suen
 * ***************************************************************************
 */
VPUBLIC int MY_bcgstab(Bvec *u, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key,
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[]){

    Bvec *r0=w[1], *pvec=w[2], *svec=w[3], *tmpvec=w[4], *tvec=w[5], *vvec=w[6];
    Bvec *z=w[7], *tmp1=w[8], *tmp2=w[9];

    strncpy(rmeth,"JKS_bcgstab",80);
    if (MY_initSolver(r, A, u, f, prec, flag, etol)) return 0;
    omega = 1.; rho=1;

    /* DO: r^0 = r */
    MY_precond(r0,A,r,P,tmp1,tmp2,prec,key);

    /* DO: for i=0,1,2,... */
    while ((!Vnm_sigInt()) && (iters < itmax)){

        /* STEP 1: pvec = r + beta*(pvec-omega*vvec) */
        beta  = alpha/(omega*rho);
        rho   = Bvec_dot(r0, r);
        beta *= rho;
        Bvec_axpy(pvec, vvec, -omega);
        Bvec_scal(pvec, beta);
        Bvec_axpy(pvec, r, 1.0);

        /* STEP 2: vvec = A * M^{-1} * pvec             */
        MY_precond(tmpvec,A,pvec,P,tmp1,tmp2,prec,key);
        Bvec_matvec(vvec,A,tmpvec,0,0);

        /* STEP 3: alpha = rho/(r0, vvec)               */
        /*         Update u = u + alpha * M^{-1} * pvec */
        alpha = rho/Bvec_dot(r0,vvec);
        if (VABS(alpha) < VSMALL){
            Vnm_print(0,"Breaking down with Alpha = %e \n", alpha);
            return iters;
        }
        Bvec_axpy(u,tmpvec,alpha);

        /* STEP 4: svec = r - alpha * vvec */
        /*         Converged if Bvec_nrm2(svec) < ETOL */
        Bvec_copy(svec,r);
        Bvec_axpy(svec,vvec,-alpha);
        if (Bvec_nrm2(svec) < mytol) return iters;

        /* STEP 5: tvec = A * M^{-1} * svec */
        MY_precond(tmpvec,A,svec,P,tmp1,tmp2,prec,key);
        Bvec_matvec(tvec,A,tmpvec,0,0);

        /* STEP 6: omega = (tvec, svec)/(tvec, tvec) */
        /*         Update u = u + omega * M^{-1} * svec */
        omega = Bvec_dot(tvec,svec)/Bvec_dot(tvec,tvec);
        if (VABS(omega) < VSMALL){
            Vnm_print(0,"Breaking down with Omega = %e \n", omega);
            return iters;
        }
        Bvec_axpy(u,tmpvec,omega);

        /* STEP 7: r = svec - omega *tvec */
        Bvec_copy(r,svec);
        Bvec_axpy(r,tvec,-omega);

        /* calculate norm of current residual, setup for i/o */
        if (MY_io(r,flag)) return iters;
        iters++;
    }
    Vnm_print(0,"%s itmax reached! (rr=%e)\n", rn, rsnrm/denom);
    return iters;
}

