/* ///////////////////////////////////////////////////////////////////////////
//  MC = < Manifold Code >
//  Copyright (C) 1994--2000  Michael Holst
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  Free Software Foundation; either version 2 of the License, or (at your
//  option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  675 Mass Ave, Cambridge, MA 02139, USA.
//
//  rcsid="$Id: mypde.h,v 1.1 2003/03/08 00:13:48 fetk Exp $"
// //////////////////////////////////////////////////////////////////////// */

#include "valist.h"
#include "vmgrid.h"
#define e_c 4.803242384e-10
#define k_B 1.380662000e-16
#define N_A 6.022045000e+23
#define kT_kcal 332.0716
#define ang3_iM 6.023e-4
#define refineParm 1.2e-5

enum Vbnd_types {
    INTERIOR=0, /**< Interior type (i.e., not a boundary point) */
    IB_RD=1,    /**< Inner boundary (IB) reactive Dirichlet (RD). */
    IB_RR=2,    /**< Inner boundary (IB) reactive Robin (RR). */
    IB_ND=3,    /**< Inner boundary (IB) non-reactive Diriclet (ND). */
    IB_NR=4,    /**< Inner boundary (IB) non-reactive Robin (ND). */
    OB_D=5,     /**< Outer boundary (OB) Dirichlet (D). */
    OB_R=6,     /**< Outer boundary (OB) Robin (R). */
};

enum METHtype {
    BEM_TYPE = 0,
    FEM_TYPE = 1,
};

typedef struct Vmodel{
    int      map, pottype, entropy, gibbs;
    int      color[15],nsp, diffmesh, tmkey;
    double   conc, charge, diff, T;
    double rad[15],ctx[15],cty[15],ctz[15];
    double *PMF, *dPMF[3], *ddPMF[3];
    Valist   *alist;
    Vmgrid   *mgrid;
} Vmodel;

typedef struct Vlparam{
    int prob,lkey,lmax,gues,prec,pjac,ptype,key,flag,cycle,tmkey;
    double ltol;
    char ifnam[VMAX_ARGLEN], ofnam[VMAX_ARGLEN], afnam[VMAX_ARGLEN];
    char Sfnam[VMAX_ARGLEN],Gfnam[VMAX_ARGLEN];
} Vlparam;

typedef struct Vtparam{
    int tkey,nstep,pfreq,efreq,cfreq,ekeytol,pdetype;
    char rstnam[VMAX_ARGLEN];
    double dt,t0;
} Vtparam;

typedef struct Vusr{
    Vmodel *model;
    Vlparam *lparam;
    Vtparam *tparam;
} Vusr;

#   define Vloc_num(thee)           (Vset_num(thee->vals))
#   define Vloc_get(thee,i)         ((Vloc*)Vset_access(thee->vals,i))
#   define Vloc_create(thee)        ((Vloc*)Vset_create(thee->vals))
#   define Vloc_first(thee)         ((Vloc*)Vset_first(thee->vals))
#   define Vloc_last(thee)          ((Vloc*)Vset_last(thee->vals))
#   define Vloc_next(thee)          ((Vloc*)Vset_next(thee->vals))
#   define Vloc_prev(thee)          ((Vloc*)Vset_prev(thee->vals))
#   define Vloc_destroy(thee)       (Vset_destroy(thee->vals))
#   define Vloc_reset(thee)         (Vset_reset(thee->vals))

#   define usr_getModel(thee)       ((Vusr*)thee->user)->model
#   define usr_getLparam(thee)      ((Vusr*)thee->user)->lparam
#   define usr_getTparam(thee)      ((Vusr*)thee->user)->tparam

VEXTERNC PDE* myPDE_ctor();
VEXTERNC Vmodel* model_ctor();
VEXTERNC Vlparam* lparam_ctor();
VEXTERNC Vtparam* tparam_ctor();
VEXTERNC Vusr* usr_ctor();

VEXTERNC void myPDE_dtor(PDE *thee);
VEXTERNC void model_dtor(Vmodel *thee);
VEXTERNC void lparam_dtor(Vlparam *thee);
VEXTERNC void tparam_dtor(Vtparam *thee);
VEXTERNC void usr_dtor(Vusr *thee);

VEXTERNC int PDE_input(AM *thee, Vgrid *grid[], Vlparam *lparam);
VEXTERNC void dPMF_mapping(AM *thee);

VEXTERNC void MY_assemR(AM *thee, Bvec *u);
VEXTERNC void MY_assem(AM *thee, int evalKey, Bvec *u, Bvec *f, int ip[], double rp[],
  double energies[], int ekeytol);
VEXTERNC int MY_Solve(AM *thee, Bvec *u, Bvec *w[], Vlparam *lparam);
VEXTERNC void MY_lSolve(AM *thee, Vlparam *lparam);
VEXTERNC void MY_tSolve(AM *thee, Vlparam *lparam, Vtparam *tparam);

VEXTERNC void MY_cont(AM *thee, Vlparam *lparam);
VEXTERNC void MY_write(AM *thee, Vlparam *lparam, double *energies);

VEXTERNC int MY_cg(Bvec *thee, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key,
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[]);

VEXTERNC int MY_cgT(Bvec *thee, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key,
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[]);
VEXTERNC int MY_bcg(Bvec *thee, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key,
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[]);
VEXTERNC int MY_bcgstab(Bvec *thee, Bmat *A, Bvec *f, Bvec *r, Bvec *ut, int key,
    int flag, int itmax, double etol, int prec, int cycle, Bmat *P, Bvec *w[]);

VEXTERNC int Vmgrid_eval(Vmgrid *thee, double pt[3], double *value,
            double grad[3], double curv[3]);

VEXTERNC int Vmgrid_curv(Vmgrid *thee, double pt[3], double curv[3]);

VEXTERNC int MY_readActive(AM *thee, Vio *sock);

VEXTERNC int MY_markRefine(Gem *thee, int key, int color);

VEXTERNC void Mesh_ActiveSite(AM *thee);

