/**
 * @defgroup  mcsf2off  MCSF-to-OFF mesh format converter
 * @brief Converts MCSF-format meshes to Geomview OFF format
 */

/**
 *  @file mcsf2off.h
 *  @ingroup mcsf2off
 *  @author  Nathan Baker
 *  @brief  Converts MCSF-format meshes to Geomview OFF format
 *  @version $Id: mcsf2off.h,v 1.1.1.1 2003/06/24 15:57:11 smol Exp $
 */

#ifndef _MCSF2OFF_H_
#define _MCSF2OFF_H_

#include "mc/mc.h"  

/**
 * @brief  Provides usage information and exits
 * @author Nathan Baker
 * @ingroup mcsf2off
 * @param  rc code to pass to exit() 
 * @notes  This function requires that Vio_start() has already been
 * called */
VPRIVATE void usage(int rc);

/**
 * @brief  Get command-line options
 * @author Nathan Baker
 * @ingroup mcsf2off
 * @param  argc number of arguments
 * @param  argv list of arguments
 * @param  mcsf_path set to input path for MCSF file
 * @param  doCheck set to 1 if mesh-checking requested
 * @param  fullMesh set to 1 if full mesh to be written out; set to 0 if only
 * boundaries to be written out
 * @returns 1 if parsed correctly, 0 otherwise
 * @notes  This function requires that Vio_start() has already been
 * called */
VPRIVATE int getopt(int argc, char **argv, char mcsf_path[VMAX_ARGLEN],
    int *doCheck, int *fullMesh);

/**
 * @brief  Provides information about program
 * @author Nathan Baker
 * @ingroup mcsf2off
 * @notes  This function requires that Vio_start() has already been
 * called */
VPRIVATE void header();

#endif /* #ifndef _MCSF2OFF_H_ */
