/** 
 * @file mypde.c
 * @brief  Contains PDE definitions for FEtk
 * @version @version $Id: mypde.c,v 1.1.1.1 2003/06/24 15:57:11 smol Exp $
 * @author   Nathan A. Baker
 */

//
// Purpose:  Provides methods for PDE class.  Does nothing.
//
// Author:   Nathan Baker
/////////////////////////////////////////////////////////////////////////// */

#include "smol.h"

/* Embed version information in object */
VEMBED(rcsid="$Id: mypde.c,v 1.1.1.1 2003/06/24 15:57:11 smol Exp $")

/* Unused functions */
VPUBLIC void initElement(PDE *thee, int elementType, int chart,
  double tvx[][3], void *data) { ; }
VPUBLIC void initFace(PDE *thee, int faceType, int chart, double tnvec[]) { ; }
VPUBLIC void initPoint(PDE *thee, int pointType, int qp, int chart, double txq[],
    double phi[], double phix[][3], double tU[], double tdU[][3]) { ; }
VPUBLIC void Fu(PDE *thee, int key, double F[]) { ; }
VPUBLIC double Fu_v(PDE *thee, int key, double V[], double dV[][3]) { ; }
VPUBLIC double DFu_wv(PDE *thee, int key, double W[], double dW[][3],
    double V[], double dV[][3]) { ; }
VPUBLIC void u_D(PDE *thee, int type, int chart, double txq[], double F[]) { ; }
VPUBLIC void u_T(PDE *thee, int type, int chart, double txq[], double F[]) { ; }
VPUBLIC void mapBoundary(int dim, int dimII, int vertexType, int chart, 
  double vx[3]) { ; }
VPUBLIC void initAssemble(PDE *thee, int ip[], double rp[]) { }
VPUBLIC void oneChart(int dim, int dimII, int objType, int chart[], 
  double vx[][3], int dimV) {}
VPUBLIC void delta(PDE *thee, int type, int chart, double txq[], void *user,
    double F[]) { ; }
VPUBLIC int markSimplex(int dim, int dimII,
    int simplexType, int faceType[4], int vertexType[4],
    int chart[], double vx[][3], void *simplex) { return 1; }

/* Geometry functions */
VPUBLIC void bisectEdge(int dim, int dimII, int edgeType, int chart[], 
  double vx[][3]) {
    int i;
    for (i=0; i<dimII; i++) vx[2][i] = .5 * (vx[0][i] + vx[1][i]);
}
VPUBLIC PDE* myPDE_ctor(void) {
    PDE *thee = VNULL;

    int i;

    /* create some space for the pde object */
    thee = Vmem_malloc(VNULL, 1, sizeof(PDE) );

    /* PDE-specific parameters and function pointers */
    thee->initAssemble = initAssemble;  /* once-per-assembly initialization */
    thee->initElement  = initElement;   /* once-per-element initialization  */
    thee->initFace     = initFace;      /* once-per-face initialization     */
    thee->initPoint    = initPoint;     /* once-per-point initialization    */
    thee->Fu           = Fu;            /* nonlinear strong form            */
    thee->Fu_v         = Fu_v;          /* nonlinear weak form              */
    thee->DFu_wv       = DFu_wv;        /* bilinear linearization weak form */
    thee->delta        = delta;         /* delta function source term       */
    thee->u_D          = u_D;           /* dirichlet func and initial guess */
    thee->u_T          = u_T;           /* analytical soln for testing      */
    thee->vec          = 1;             /* unknowns per spatial point;      */
    thee->sym[0][0]    = 0;             /* symmetries of bilinear form      */
    thee->est[0]       = 1.0;           /* error estimator weights          */
    for (i=0; i<VMAX_BDTYPE; i++)       /* boundary type remappings         */
        thee->bmap[0][i] = i;

    /* Manifold-specific function pointers */
    thee->bisectEdge  = bisectEdge;  /* edge bisection rule                 */
    thee->mapBoundary = mapBoundary; /* boundary recovery rule              */
    thee->markSimplex = markSimplex; /* simplex marking rule                */
    thee->oneChart    = oneChart;    /* coordinate transformations          */

    /* Element-specific function pointers */
//    thee->simplexBasisInit = simplexBasisInit; /* initialization of bases   */
//    thee->simplexBasisForm = simplexBasisForm; /* form trial & test bases   */

    /* return the new pde object */
    return thee;
}
VPUBLIC void myPDE_dtor(PDE *thee) {

    VASSERT( (thee) != VNULL );
    if ((thee) != VNULL) {
        Vmem_free(VNULL, 1, sizeof(PDE), (void **)&thee);
        thee = VNULL;
    }
}

