#undef _DEBUG

/* symbolic encoding of method names */
VPRIVATE char *rnt[4] = {
   "AM_fEuler",    /* 0 */
   "AM_bEuler",    /* 1 */
   "AM_tSICN ",    /* 2 */
   "AM_tRK4  "     /* 3 */
};

VPRIVATE double iorder[4] = { /* Order of the method per substep */
    1.0,    /* Forward Euler */
    1.0,    /* Backward Euler */
    2.0,    /* Crank Nicholson */
    1.0     /* RK4 */
};

VPRIVATE double MY_anal(double radius, double dtm, int iter, double dif){

    double sincoef,expcoef[20], analval, analcoef = 0.0;
    double tm;
    int i;

    sincoef = VPI*radius/50.0;
    if (iter > 0) {
    tm = 0 + dtm*iter;
    for(i=1;i<40;i++) {
        expcoef[i] = VPOW(-1,i)*VSIN(i*sincoef)*VEXP(-VSQR(VPI*i/50.0)*dif*tm)/i;
        analcoef += expcoef[i];
    }
    analval = 1.0 + (2.0/sincoef)*analcoef;
    }else {
    analval = 0.0;
    }
    return analval;
}

VPUBLIC void MY_tSolve(AM *thee, Vlparam *lparam, Vtparam *tparam){

    int meth=tparam->tkey, numstep=tparam->nstep, pdetype=tparam->pdetype;
    int pfreq=tparam->pfreq, efreq=tparam->efreq, ekeytol=tparam->ekeytol;
    int cfreq=tparam->cfreq;
    double dt=tparam->dt, t0=tparam->t0;

    int i, istep, rstep, piters=0, ekey, ip[20], iteru=0, imadeVlines=0;
    int evalKey=1, energyKey=0, residKey=1, tangKey=1, massKey=0;
    int numB, numR[MAXV];
    double energy, energies[20], rp[20], tcpu, y[3], rad, theta = 1.0, time = t0;
    double energies0=0.0,subnum=0.0, rstcon;
    Bvec *u=thee->u, *ud=thee->ud, *f=thee->f, *w0=thee->w0;
    Bvec *deltu, *resid, *ut=thee->ut, *w[len_w];
    Bmat *A = thee->A, *M = thee->M, *P = thee->P, *cmat;
    SVio *vsock = VNULL;
    Vio *sock=VNULL;
    Aprx *aprx= thee->aprx;
    Gem  *gm  = aprx->gm;
    PDE *pde = thee->aprx->pde;
    Vmodel *model = usr_getModel(pde);
    VV *vx;
    SS *sm;
    TT t;
    char tok[VMAX_BUFSIZE];
    char *MCwhiteChars = " =,;\t\n", *MCcommChars  = "#%";
    FILE *fp;

    Vio_start();

    AM_create(thee);

    vsock = SVio_ctor();
    SVio_initStructure(vsock, "w", "FILE", "ASC", "",
                    lparam->ofnam, VNULL, 0, lparam->ptype);

    if (meth == 0) pdetype = 0;

    Vnm_tstart(41, "Loop");

    Vnm_print(0,"AM_tSolve: starting..");

    /* Initialize work vectors */
    numB = u->numB;
    for (i=0; i<numB; i++) numR[i] = Bvec_numRB(u,i);

    Bvec_createVectors( u, w, len_w );
    deltu = Bvec_ctor(VNULL,"deltu", numB, numR);
    resid = Bvec_ctor(VNULL,"resid", numB, numR);
    ut    = w[0];

    AM_iniGuess( thee, u );         /* setup initial condition */

    fp = fopen(tparam->rstnam,"r");
    if (fp != (FILE *) NULL) {
       printf("Opening %s.\n",tparam->rstnam);
       sock = Vio_ctor("FILE", "ASC", "localhost", tparam->rstnam, "r");
       Vio_setWhiteChars(sock, MCwhiteChars);
       Vio_setCommChars(sock, MCcommChars);
       VJMPERR2( 1==Vio_scanf(sock,"%s",tok) );
       VJMPERR2( !strcmp(tok,"u_0") );
       VJMPERR2( 1==Vio_scanf(sock,"%s",tok) );
       VJMPERR2( !strcmp(tok,"[") );
       for (i=0; i<Gem_numVV(gm); i++) {
           VJMPERR2( 1==Vio_scanf(sock,"%le",&rstcon) );
           Bvec_setB(u,0,i,rstcon);
       }
       VJMPERR2( 1==Vio_scanf(sock,"%s",tok) );
       VJMPERR2( !strcmp(tok,"]") );
       Vio_dtor(&sock);
       fclose(fp);
    }

    rp[0] = 1/dt;
    rp[1] = 0.0;
    if (pdetype>0) rp[1] = 1.0;

    /* some i/o */
    Vnm_print(0,"AM_tSolve: starting..");

    Bmat_zero( A );
    Bmat_zero( M );
    Bvec_init( f, 0. );
    Bvec_init( w0, 0. );

    /* assemble the stiffness matrix and load vector  */
    rp[0] = 0.;
    Aprx_assem(thee->aprx, evalKey, energyKey, residKey, tangKey, massKey, 0,
                ip, rp, M, A, f, u, ud);
    rp[0] = 1/dt;
    Aprx_assem(thee->aprx, evalKey, energyKey, residKey, tangKey, massKey, 0,
                ip, rp, A, M, f, u, ud);
   
    for (i=0; i<ekeytol; i++) ip[i] = i;
    
    Vnm_tstart(40, "time solve");
    Vnm_print(0,"%s: taking %d steps with dt = %12.8e.\n",
        rnt[meth], numstep, dt);

    Vnm_print(2,"     time ");
    for (i=0; i<ekeytol; i++) Vnm_print(2,"      Ju(%1d)", i);
    Vnm_print(2," iters \n");

    rstep = ceil(t0/dt);
    for (istep=0; istep<= numstep; istep++){    /* Start Main Loop */

        if (Vnm_sigInt()) goto DONE;

        if ( (pfreq > 0) && !(istep%pfreq) && (istep > 0)) {
            sprintf(tparam->rstnam,"rst.m.%d",istep+rstep);
            Bvec_printSp(u,tparam->rstnam);
        }

        if (efreq > 0) {       
            if ( !(istep%efreq) ) {
               Vnm_print(2, " %10.6f", t0+istep*dt);
               if (istep == 0) {
                  MY_assem(thee,0,u,f,ip,rp,energies,ekeytol); 
                  energies0 = energies[2];
               }
               MY_assem(thee,1,u,f,ip,rp,energies,ekeytol);
               energies[0] = energies[0]/model->conc;
               subnum = energies[2] - energies0;
               Vnm_print(2," %17.5e  %17.5e  %10.5e  %10.5e",energies[0],
                  energies[0]*6e7*ang3_iM,subnum,subnum/energies[3]);
               Vnm_print(2, " %6d \n", iteru);
            }
            if ( !(istep%cfreq) ) {
               sprintf(lparam->ofnam,"conc.dx.%d",istep+rstep);
               MY_write(thee,lparam, energies);                      
            }
        }

        Vnm_tstart(42,"assem");
        Bvec_init(deltu, 0.0);
        Bvec_matvec(f, M, u, 0, 0);
        Bvec_scal(f, -1.0);
        Aprx_assem( thee->aprx, 0, 0, 1, 0, 0, 0, ip, rp, A, M, f, u, ud);
        Vnm_tstop(42,"assem");
        Vnm_tstart(42,"solve");
        iteru   = MY_Solve(thee, deltu, w, lparam);
        Bvec_axpy(u, deltu, iorder[meth]);
        Vnm_tstop(42,"solve");
   }                          

DONE:
    Vnm_tstop(40, "time solve");

    Vnm_print(0,"AM_tSolve: done.\n");
        
    Bvec_destroyVectors( u, w, len_w );
    Bvec_dtor( &(deltu) );
    Bvec_dtor( &(resid) );
    SVio_dtor(&vsock);

    return;                     /* return w/o errors */

VERROR2:
  Vnm_print(2,"RST_read: Format problem with input\n");
  Vnm_print(2,"RST_read: Reseting manifold structures (bailing out).\n");
  Vio_dtor(&sock);
  return;

}

VPUBLIC Vtparam* tparam_ctor(){
    Vtparam *thee;
    thee = Vmem_malloc( VNULL, 1, sizeof(Vlparam) );
    return thee;
}

VPUBLIC void tparam_dtor(Vtparam *thee){
    Vmem_free(VNULL, 1, sizeof(Vtparam), (void **)(&(thee)));
}

